/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 * @emails oncall+relay
 * @format
 */
// flowlint ambiguous-object-type:error
'use strict';

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _require = require('./RelayReactCache'),
    getCacheForType = _require.getCacheForType,
    getCacheSignal = _require.getCacheSignal;

var invariant = require('invariant');

var _require2 = require('relay-runtime'),
    fetchQueryInternal = _require2.__internal.fetchQuery;

var warning = require("fbjs/lib/warning");

var DEFAULT_FETCH_POLICY = 'store-or-network';

function createQueryCache() {
  return new Map();
}

function getQueryCacheKey(operation, fetchPolicy, renderPolicy, fetchKey) {
  var cacheIdentifier = "".concat(fetchPolicy, "-").concat(renderPolicy, "-").concat(operation.request.identifier, "-").concat(fetchKey !== null && fetchKey !== void 0 ? fetchKey : '');
  return cacheIdentifier;
}

function constructQueryResult(operation) {
  var rootFragmentRef = {
    __id: operation.fragment.dataID,
    __fragments: (0, _defineProperty2["default"])({}, operation.fragment.node.name, operation.request.variables),
    __fragmentOwner: operation.request
  };
  return {
    fragmentNode: operation.request.node.fragment,
    fragmentRef: rootFragmentRef
  };
}

function getQueryResultOrFetchQuery_REACT_CACHE(environment, queryOperationDescriptor, options) {
  var _options$fetchPolicy, _options$renderPolicy;

  var fetchPolicy = (_options$fetchPolicy = options === null || options === void 0 ? void 0 : options.fetchPolicy) !== null && _options$fetchPolicy !== void 0 ? _options$fetchPolicy : DEFAULT_FETCH_POLICY;
  var renderPolicy = (_options$renderPolicy = options === null || options === void 0 ? void 0 : options.renderPolicy) !== null && _options$renderPolicy !== void 0 ? _options$renderPolicy : environment.UNSTABLE_getDefaultRenderPolicy();
  var cache = getCacheForType(createQueryCache);
  var cacheKey = getQueryCacheKey(queryOperationDescriptor, fetchPolicy, renderPolicy, options === null || options === void 0 ? void 0 : options.fetchKey);
  var entry = cache.get(cacheKey);

  if (entry === undefined) {
    // Initiate a query to fetch the data if needed:
    entry = onCacheMiss(environment, queryOperationDescriptor, fetchPolicy, renderPolicy, function (newCacheEntry) {
      cache.set(cacheKey, newCacheEntry);
    }, options === null || options === void 0 ? void 0 : options.fetchObservable);
    cache.set(cacheKey, entry); // Since this is the first time rendering, retain the query. React will
    // trigger the abort signal when this cache entry is no longer needed.

    var retention = environment.retain(queryOperationDescriptor);
    var abortSignal = getCacheSignal();
    abortSignal.addEventListener('abort', function () {
      retention.dispose();
      cache["delete"](cacheKey);
    }, {
      once: true
    });
  }

  switch (entry.status) {
    case 'pending':
      throw entry.promise;

    case 'rejected':
      throw entry.error;

    case 'resolved':
      return entry.result;
  }

  !false ? process.env.NODE_ENV !== "production" ? invariant(false, 'switch statement should be exhaustive') : invariant(false) : void 0;
}

function onCacheMiss(environment, operation, fetchPolicy, renderPolicy, updateCache, customFetchObservable) {
  // NB: Besides checking if the data is available, calling `check` will write missing
  // data to the store using any missing data handlers specified in the environment.
  var queryAvailability = environment.check(operation);
  var queryStatus = queryAvailability.status;
  var hasFullQuery = queryStatus === 'available';
  var canPartialRender = hasFullQuery || renderPolicy === 'partial' && queryStatus !== 'stale';
  var shouldFetch;
  var shouldRenderNow;

  switch (fetchPolicy) {
    case 'store-only':
      {
        shouldFetch = false;
        shouldRenderNow = true;
        break;
      }

    case 'store-or-network':
      {
        shouldFetch = !hasFullQuery;
        shouldRenderNow = canPartialRender;
        break;
      }

    case 'store-and-network':
      {
        shouldFetch = true;
        shouldRenderNow = canPartialRender;
        break;
      }

    case 'network-only':
    default:
      {
        shouldFetch = true;
        shouldRenderNow = false;
        break;
      }
  }

  var cacheEntry;

  if (shouldFetch) {
    cacheEntry = executeOperationAndKeepUpToDate(environment, operation, updateCache, customFetchObservable);
  }

  if (cacheEntry) {
    switch (cacheEntry.status) {
      case 'resolved':
        return cacheEntry;

      case 'rejected':
        return cacheEntry;

      case 'pending':
        return shouldRenderNow ? {
          status: 'resolved',
          result: constructQueryResult(operation)
        } : cacheEntry;
    }
  } else {
    !shouldRenderNow ? process.env.NODE_ENV !== "production" ? invariant(false, 'Should either fetch or be willing to render. This is a bug in Relay.') : invariant(false) : void 0;
    return {
      status: 'resolved',
      result: constructQueryResult(operation)
    };
  }
}

function executeOperationAndKeepUpToDate(environment, operation, updateCache, customFetchObservable) {
  var _entry;

  var resolvePromise;
  var promise = new Promise(function (r) {
    resolvePromise = r;
  }); // $FlowExpectedError[prop-missing] Expando to annotate Promises.

  promise.displayName = 'Relay(' + operation.request.node.operation.name + ')';
  var isFirstPayload = true;
  var entry; // FIXME We may still need to cancel network requests for live queries.

  var fetchObservable = customFetchObservable !== null && customFetchObservable !== void 0 ? customFetchObservable : fetchQueryInternal(environment, operation);
  fetchObservable.subscribe({
    start: function start(subscription) {},
    error: function error(_error) {
      if (isFirstPayload) {
        entry = {
          status: 'rejected',
          error: _error
        };
        updateCache(entry);
      } else {
        // TODO:T92030819 Remove this warning and actually throw the network error
        // To complete this task we need to have a way of precisely tracking suspendable points
        process.env.NODE_ENV !== "production" ? warning(false, 'getQueryResultOrFetchQuery: An incremental payload for query `%` returned an error: `%`:`%`.', operation.request.node.operation.name, _error.message, _error.stack) : void 0;
      }

      resolvePromise();
      isFirstPayload = false;
    },
    next: function next(response) {
      // Stop suspending on the first payload because of streaming, defer, etc.
      entry = {
        status: 'resolved',
        result: constructQueryResult(operation)
      };
      updateCache(entry);
      resolvePromise();
      isFirstPayload = false;
    },
    complete: function complete() {
      // FIXME I don't think we need to do anything further on complete.
      entry = {
        status: 'resolved',
        result: constructQueryResult(operation)
      };
      updateCache(entry);
      resolvePromise();
      isFirstPayload = false;
    }
  });
  return (_entry = entry) !== null && _entry !== void 0 ? _entry : {
    status: 'pending',
    promise: promise
  };
}

module.exports = getQueryResultOrFetchQuery_REACT_CACHE;